/**
 * Created on 2008-06-12 12:15:21
 */
package org.put.semintec.mining.association;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import jdsl.core.api.Position;
import jdsl.core.api.PositionIterator;

import org.put.semintec.core.KnowledgeBase;
import org.put.semintec.mining.datastructures.Node;
import org.put.semintec.mining.datastructures.Trie;
import org.semanticweb.kaon2.api.KAON2Exception;
import org.semanticweb.kaon2.api.OntologyChangeEvent;

/**
 * Represents an algorithm for generation of relational association rules (query extensions) from 
 * discovered frequent patterns.
 * @author Agnieszka Lawrynowicz
 * 
 */
public class AssociationRuleGeneration {
	/**
	 * Trie data structure with discovered frequent patterns on which rule generation algorithm operates.
	 */
	private Trie trie;
	 
	/**
	 * Getter of the property <tt>trie</tt>
	 *
	 * @return Returns the trie.
	 * 
	 */
	public Trie getTrie()
	{
		return trie;
	}

	/**
	 * Setter of the property <tt>trie</tt>
	 *
	 * @param trie The trie to set.
	 *
	 */
	public void setTrie(Trie trie ){
		this.trie = trie;
	}
	
	
	/**
	 * Copy of the knowledge base on which rule generation algorithm is run.
	 * Used in containment tests
	 */
	private KnowledgeBase kbCopy;

	/**
	 * Setter of the property <tt>kbCopy</tt>
	 *
	 * @param kbCopy The kbCopy to set.
	 *
	 */
	public void setKbCopy(KnowledgeBase kbCopy ){
		this.kbCopy = kbCopy;
	}

	 
	/**
	 * Getter of the property <tt>kbCopy</tt>
	 *
	 * @return Returns the kbCopy.
	 * 
	 */
	public KnowledgeBase getKbCopy()
	{
		return kbCopy;
	}

	
	/**
	 * List of generated rules
	 */
	private ArrayList rules;
	 
	/**
	 * Getter of the property <tt>rules</tt>
	 *
	 * @return Returns the rules.
	 * 
	 */
	public ArrayList getRules()
	{
		return rules;
	}

	/**
	 * Setter of the property <tt>rules</tt>
	 *
	 * @param rules The rules to set.
	 *
	 */
	public void setRules(ArrayList rules){
		this.rules = rules;
	}
	
	/**
	 * Initializes rule generation algorithm 
	*/
	private void initialize() {
		rules = new ArrayList();
	}
	
	/**
	 * Executes the algorithm generating rules. 
	 */
	public void generate(){
        initialize();

		try {
			breadthFirstExpand((Node)trie.getTree().root().element(), kbCopy);
		} catch (KAON2Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	/**
	 * Generates rules expanding the trie breadth first 
	 *  
	 * @param n Current node to expand
	 * @param kb Knowledge base wrt which rule redundancy is checked
	 * @throws InterruptedException 
	 * @throws KAON2Exception 
	*/

	private void breadthFirstExpand(Node n, KnowledgeBase kb) throws KAON2Exception, InterruptedException{
		Position parent = n.getPos();
		if (trie.getTree().isInternal(n.getPos())) {
			List<OntologyChangeEvent> rollback = null;
			
			//iterate over children of given node to create rules
			for (PositionIterator i = trie.getTree().children(n.getPos()); i.hasNext();) {
				//take next child
				Position pChild = i.nextPosition();
				Node nChild = (Node) pChild.element();
				boolean reachedRoot = false;
				
				AssociationRule ar = new AssociationRule();
				ar.setAntecedent(n.getPattern());
				ar.setConsequent(nChild.getPattern());
				ar.calculateConfidence();
				if (!redundantWrtEarlierRule(ar, kb)){
					rules.add(ar);
				}
				
				Position ancestor = null;
				if (!trie.getTree().isRoot(parent)){
					ancestor = trie.getTree().parent(parent); 
				} else {
					reachedRoot = true;
				}
				while (!reachedRoot){
					ar = new AssociationRule();
					ar.setAntecedent(((Node)ancestor.element()).getPattern());
					ar.setConsequent(nChild.getPattern());
					ar.calculateConfidence();
					if (!redundantWrtEarlierRule(ar, kb)){
						rules.add(ar);
					}
					if (!trie.getTree().isRoot(ancestor)){
						ancestor = trie.getTree().parent(ancestor); 
					} else {
						reachedRoot = true;
					}
				}
			}
			
			for (PositionIterator i = trie.getTree().children(n.getPos()); i.hasNext();) {
				//take next child
				Position pChild = i.nextPosition();
				Node nChild = (Node) pChild.element();
				breadthFirstExpand(nChild, kb);
			}
		}
	}
	
	/**
	 * Checks if the specified rule has a consequent that subsumes a consequent 
	 * of an already generated rule with the same antecedent as that of a given one.
	 * If so, then the given rule is called redundant, and there is already generated essential one.
	 * Check is done by searching the list of already generated rules. 
	 * @param r Rule whose essential rule is searched in the trie
	 * @param kb Knowledge base wrt which subsumption/containmnet of rule consequents is checked
	 * @return True if the rule is redundant due to already generated one, false otherwise 
	* @throws InterruptedException 
	* @throws KAON2Exception 
	*/
	public boolean redundantWrtEarlierRule(AssociationRule rule, KnowledgeBase kb) throws KAON2Exception, InterruptedException{
		boolean redundant = false;
		Iterator i = getRules().iterator();
		while (!redundant && i.hasNext()){
			AssociationRule r = (AssociationRule) i.next();
			if (rule.getAntecedent().equivalentTo(r.getAntecedent(), kb)) {
				if (r.getConsequent().subsumedBy(rule.getConsequent(), kb)) {
					redundant = true;
				}
			}
		}
		return redundant;
	}
}
