/**
 * Created on 2008-03-28 14:37:53
 */
package org.put.semintec.mining.association;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import jdsl.core.api.Position;
import jdsl.core.api.Tree;
import jdsl.core.ref.NodeTree;

import org.put.semintec.core.KnowledgeBase;
import org.semanticweb.kaon2.api.KAON2Exception;
import org.semanticweb.kaon2.api.KAON2Manager;
import org.semanticweb.kaon2.api.logic.Predicate;
import org.semanticweb.kaon2.api.owl.elements.OWLClass;
import org.semanticweb.kaon2.api.owl.elements.ObjectProperty;
import org.semanticweb.kaon2.api.reasoner.SubsumptionHierarchy;


/**
 * Represents a declarative bias.
 * @author Agnieszka Lawrynowicz
 * 
 */
public class Bias {

	public Bias(){
		predicates = new ArrayList();
		classTaxonomy = new NodeTree();
		propertyTaxonomy = new NodeTree();
	}

	/**
 * The list of predicates from which patterns are to be generated.
 */
public static ArrayList predicates;
 
/**
 * Getter of the property <tt>predicates</tt>
 *
 * @return Returns the predicates.
 * 
 */
public static ArrayList getPredicates()
{
	return predicates;
}

/**
 * Setter of the property <tt>predicates</tt>
 *
 * @param predicates The predicates to set.
 *
 */
public void setPredicates(ArrayList predicates ){
	this.predicates = predicates;
}

/**
 * Class predicates given in declarative bias, arranged into taxonomy due to classified KB
 */
private Tree classTaxonomy;
 
/**
 * Getter of the property <tt>classTaxonomy</tt>
 *
 * @return Returns the classTaxonomy.
 * 
 */
public Tree getClassTaxonomy()
{
	return classTaxonomy;
}

/**
 * Setter of the property <tt>classTaxonomy</tt>
 *
 * @param classTaxonomy The classTaxonomy to set.
 *
 */
public void setClassTaxonomy(Tree classTaxonomy){
	this.classTaxonomy = classTaxonomy;
}


/**
 * Property predicates given in declarative bias, arranged into taxonomy
 */
private Tree propertyTaxonomy;
 
/**
 * Getter of the property <tt>propertyTaxonomy</tt>
 *
 * @return Returns the propertyTaxonomy.
 * 
 */
public Tree getPropertyTaxonomy()
{
	return propertyTaxonomy;
}

/**
 * Setter of the property <tt>propertyTaxonomy</tt>
 *
 * @param propertyTaxonomy The propertyTaxonomy to set.
 *
 */
public void setPropertyTaxonomy(Tree propertyTaxonomy){
	this.propertyTaxonomy = propertyTaxonomy;
}

/**
 * Computes taxonomy of class predicates given in declarative bias, wrt the classified KB
 * A root of class taxonomy is owl:thing
 *
 * @param kb Knowledge base
 * 
 */
public void computeClassTaxonomy(KnowledgeBase kb){
	OWLClass thing = KAON2Manager.factory().thing();	
	buildTaxonomyRootedAtClass(thing, Bias.predicates, kb, classTaxonomy);
}

/**
 * Computes taxonomy of property predicates given in declarative bias
 * A root of property taxonomy is a fake property simulating top property
 * @param kb Knowledge base
 * 
 */
public void computePropertyTaxonomy(KnowledgeBase kb){
	ObjectProperty topFakeProperty = KAON2Manager.factory().objectProperty("$topFakeProperty$");	
	//insert top fake property  at root
	propertyTaxonomy.replaceElement(propertyTaxonomy.root(), topFakeProperty);

	ArrayList tp = selectTopPropertiesFromList(predicates, kb);
	for (Iterator i = tp.iterator(); i.hasNext();){
		ObjectProperty op = (ObjectProperty) i.next();
		Position newPos = propertyTaxonomy.insertLastChild(propertyTaxonomy.root(), op);
		recursivelyBuildPropertyTaxonomy(propertyTaxonomy, newPos, op, predicates, kb);
	}	
}

/**
 * Creates a taxonomy of classes, from the list of predicates specified as input,
 * that are rooted at the specified class according to classified knowledge base
 *
 * @param topClass Top class to be root of the taxonomy
 * @param predicates List of predicates, from which classes are to be selected
 * @param kb Knowledge base
 * 
 * @return Taxonomy of classes, rooted at specified class, from those appearing 
 * in input predicate list  
 */
public Tree buildTaxonomyRootedAtClass(OWLClass topClass, ArrayList predicates, KnowledgeBase kb, Tree taxonomy) {
	//insert top class at root
	taxonomy.replaceElement(taxonomy.root(), topClass);
	
	//get the node with topClass from subsumption hierarchy
    SubsumptionHierarchy sh;
	try {
		sh = kb.getReasoner().getSubsumptionHierarchy();
		SubsumptionHierarchy.Node topNode = sh.getNodeFor(topClass); 
		
		taxonomy = recursivelyBuildTaxonomy(taxonomy, taxonomy.root(), topNode, predicates);
		
	} catch (KAON2Exception e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	} catch (InterruptedException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}

	return taxonomy;
}


/**
 * Recursively builds a part of taxonomy of classes, from the list of predicates specified as input,
 * that is rooted at the specified parent class according to specified knowledge base
 *
 * @param taxonomy Whole taxonomy
 * @param pos Position in a taxonomy of parent class (to be root of current branch)
 * @param parent A class to be root of the part of taxonomy (that is a root of one of its branches)
 * @param predicates List of predicates, from among which to build a taxonomy
 * 
 * @return Part of the taxonomy of classes, rooted at specified parent class, 
 * build only from such classes that appear on input predicate list  
 */
Tree recursivelyBuildTaxonomy(Tree taxonomy, Position pos, SubsumptionHierarchy.Node parent, ArrayList predicates){
    Set s = parent.getChildNodes();
    OWLClass nothing = KAON2Manager.factory().nothing();
 
    Set sp = parent.getOWLClasses();
    
	//get children of parent node
    for (Iterator i=s.iterator(); i.hasNext();)   {
    	SubsumptionHierarchy.Node shn = (SubsumptionHierarchy.Node) i.next();
    	Set ss = shn.getOWLClasses();

    	for (Iterator ii=ss.iterator(); ii.hasNext();) {
    		OWLClass c = (OWLClass) ii.next();
            if (!c.equals(nothing)){ 
    	    	//check whether a class from subsumption hierarchy is on predicates list
    	    	if ((predicates.contains(c) || predicates.isEmpty()) ){
    	    		//put the class into taxonomy at the right place
    	    		Position newPos = taxonomy.insertLastChild(pos, c);
    	    		//recursively traverse hierarchy
    	    		taxonomy = recursivelyBuildTaxonomy(taxonomy, newPos, shn, predicates);
    	    	} else {
    	    		//recursively traverse hierarchy
    	    		taxonomy = recursivelyBuildTaxonomy(taxonomy, pos, shn, predicates);
    	    	}
    	    }
    	}	       
    }
    return taxonomy;
}

/**
 * Recursively builds a part of taxonomy of properties, only from the predicates specified on input,
 * that is rooted at the specified parent property according to specified knowledge base
 *
 * @param taxonomy Whole taxonomy
 * @param pos Position in a taxonomy of parent property (to be root of current branch)
 * @param parent A property to be root of the part of taxonomy (that is a root of one of its branches)
 * @param predicates List of predicates, from among which to build a taxonomy
 * @param kb KnowledgeBase
 * 
 * @return Part of the taxonomy of properties, rooted at specified parent property, 
 * build only from such properties that appear on input predicate list  
 */
Tree recursivelyBuildPropertyTaxonomy(Tree taxonomy, Position pos, ObjectProperty parent, ArrayList predicates, KnowledgeBase kb){
    Set s = null;
	try {
		s = parent.getSubObjectProperties(kb.getOntology());
	} catch (KAON2Exception e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}
    
	//get children of parent node
    for (Iterator i=s.iterator(); i.hasNext();)   {
    	ObjectProperty op = (ObjectProperty) i.next();
 	    //check whether a property is on predicates list
    	if ((predicates.contains(op) || predicates.isEmpty()) ){
    		//put the property into taxonomy at the right place
    	    Position newPos = taxonomy.insertLastChild(pos, op);
    	    //recursively traverse hierarchy
    	    taxonomy = recursivelyBuildPropertyTaxonomy(taxonomy, newPos, op, predicates, kb);
    	} else {
    	 	//recursively traverse hierarchy
    	  	taxonomy = recursivelyBuildPropertyTaxonomy(taxonomy, pos, op, predicates, kb);
    	}
    }
    return taxonomy;
}

/**
 * Retrieves all classes from the list of predicates, specified as input,
 * that have no class subsuming it present at the list of predicates
 *
 * @param predicates List of predicates, from which only classes without any subsumer are to be selected
 * @param kb Knowledge base
 * 
 * @return List of classes without subsumers, from those appearing in input predicate list  
 */
public ArrayList selectTopClassesFromList(ArrayList predicates, KnowledgeBase kb){
	ArrayList topClasses = new ArrayList(); 
	for (Iterator i = predicates.iterator(); i.hasNext();){
		Predicate p = (Predicate) i.next();
		if (p instanceof OWLClass){
			boolean noSubsumer = true;
			
			for (Iterator ii = predicates.iterator(); ii.hasNext();){
				Predicate p2 = (Predicate) ii.next();
				
				if (p2 instanceof OWLClass){
					if (!p2.equals(p)){
						try {
							if (kb.getReasoner().subsumedBy((OWLClass) p, (OWLClass)p2)){
								noSubsumer = false;
							}
						} catch (KAON2Exception e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						} catch (InterruptedException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
					}
				}
			}
			if (noSubsumer){
				topClasses.add(p);
			}
		}
	}
	
	return topClasses;	
}


/**
 * Retrieves all properties from the list of predicates, specified as input,
 * that have no super property (transitively) present at the list of predicates
 *
 * @param predicates List of predicates, from which only properties without any super property are to be selected
 * @param kb Knowledge base
 * 
 * @return List of properties without super properties, from those appearing in input predicate list  
 */
public ArrayList selectTopPropertiesFromList(ArrayList predicates, KnowledgeBase kb){
	ArrayList topProperties = new ArrayList(); 
	for (Iterator i = predicates.iterator(); i.hasNext();){
		Predicate p = (Predicate) i.next();
		if (p instanceof ObjectProperty){
			boolean noSuperProperty = true;
			
			for (Iterator ii = predicates.iterator(); ii.hasNext();){
				Predicate p2 = (Predicate) ii.next();
				
				if (p2 instanceof ObjectProperty){
					if (!p2.equals(p)){
						if (isSuperPropertyTransitively((ObjectProperty)p2, (ObjectProperty)p, kb)) {	
							noSuperProperty = false;
						}
					}
				}
			}
			if (noSuperProperty){
				topProperties.add(p);
			}
		}
	}
	
	return topProperties;	
}


/**
 * @param opSup Object property to check for being super-property of opSub,
 * by a chain of subObjectPropertyOf axioms
 * @param opSub Object property to check for being sub-property of opSup,
 * by a chain of subObjectPropertyOf axioms
 * @param kb Knowledge base
 *  
 * @return true If opSup is (transitively) super-property of opSub, false otherwise
 * */
public boolean isSuperPropertyTransitively(ObjectProperty opSup, ObjectProperty opSub, KnowledgeBase kb){
	boolean result = false;
	Set subProps = null;
	
	try {
		subProps = opSup.getSubObjectProperties(kb.getOntology());
	} catch (KAON2Exception e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}

	for (Iterator i = subProps.iterator(); i.hasNext();) {
		ObjectProperty op = (ObjectProperty) i.next();
		if (op.equals(opSub)) {
			result = true;
		}
		else {
			if (isSuperPropertyTransitively(op, opSub, kb)) {
				result = true;
			}
		}
	}
	
	return result;
}

/**
 * Retrieves classes that are top classes in given KB from the list of predicates specified as input. 
 *
 * @param predicates List of predicates, from which only top classes are to be selected
 * @param kb Knowledge base
 * 
 * @return List of top classes, from those appearing in input predicate list  
 */
public ArrayList determineTopKBClassesInList(ArrayList predicates, KnowledgeBase kb){
	ArrayList topClasses; 
	ArrayList filteredTopClasses = new ArrayList();
	
	topClasses = getTopClasses(kb);
	for (Iterator i = topClasses.iterator(); i.hasNext();){
		OWLClass c = (OWLClass) i.next();
		if (predicates.contains(c)){
			filteredTopClasses.add(c);
		}
	}
	
	return filteredTopClasses;	
}

/**
 * Retrieves all top classes from given knowledge base 
 *
 * @param kb Knowledge base
 * 
 * @return List of top classes  
 */
ArrayList getTopClasses(KnowledgeBase kb){
	ArrayList topClasses = new ArrayList();

    SubsumptionHierarchy sh;
	try {
		sh = kb.getReasoner().getSubsumptionHierarchy();
        SubsumptionHierarchy.Node thing = sh.thingNode(); 

        Set s = thing.getChildNodes();

        for (Iterator i=s.iterator(); i.hasNext();)   {
        	SubsumptionHierarchy.Node shn = (SubsumptionHierarchy.Node) i.next();
        	Set ss = shn.getOWLClasses();
        	for (Iterator ii=ss.iterator(); ii.hasNext();) {
        		OWLClass c = (OWLClass) ii.next();
        		topClasses.add(c);
        	}	       
        }
	} catch (KAON2Exception e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	} catch (InterruptedException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}

	return topClasses;	
}

}
