/**
 * Created on 2008-03-18 11:54:52
 */
package org.put.semintec.mining.datastructures;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import jdsl.core.api.Position;
import jdsl.core.api.PositionIterator;
import jdsl.core.api.Tree;
import jdsl.core.ref.NodeTree;
import org.put.semintec.core.Atom;
import org.put.semintec.core.KnowledgeBase;
import org.put.semintec.mining.association.Bias;
import org.put.semintec.mining.association.Pattern;
import org.put.semintec.utils.PermutationsWithRepetition;
import org.put.semintec.utils.IOUtils;
import org.semanticweb.kaon2.api.KAON2Exception;
import org.semanticweb.kaon2.api.KAON2Manager;
import org.semanticweb.kaon2.api.logic.Conjunction;
import org.semanticweb.kaon2.api.logic.Literal;
import org.semanticweb.kaon2.api.logic.Predicate;
import org.semanticweb.kaon2.api.logic.PredicateSymbol;
import org.semanticweb.kaon2.api.logic.Term;
import org.semanticweb.kaon2.api.logic.Variable;
import org.semanticweb.kaon2.api.owl.elements.OWLClass;
import org.semanticweb.kaon2.api.owl.elements.OWLEntity;
import org.semanticweb.kaon2.api.owl.elements.ObjectProperty;
import org.semanticweb.kaon2.api.reasoner.SubsumptionHierarchy;



/**
 * Represents a trie data structure to store patterns. 
 * Nodes represents atoms of a query. 
 * Every path from a trie root to a node/leaf represents a query.
 * @author Agnieszka Lawrynowicz
 * 
 */
public class Trie {

/**
 * Used to generate new variable names at each depth.
 * The names of new variables at each depth start from <tt>currentDepth*MAX_VARS</tt>
 */

public final static int MAX_VARS = 1000;


	/**
	 * @param n Root of the trie
	 */

	public Trie(Node n){
		tree = new NodeTree();
		n.setPos(tree.root());
		tree.replaceElement(tree.root(), n);
	}

/**
 * A tree data structure as the core of the trie one.
 */
private Tree tree;
 
/**
 * Getter of the property <tt>tree</tt>
 *
 * @return Returns the tree.
 * 
 */
public Tree getTree()
{
	return tree;
}

/**
 * Setter of the property <tt>tree</tt>
 *
 * @param tree The tree to set.
 *
 */
public void setTree(Tree tree ){
	this.tree = tree;
}

/**
 * Checks if the specified pattern is not equivalent to an earlier pattern in the trie.
 * Check is done by searching the trie. 
 * @param p Pattern whose equivalent is searched in the trie
 * @param kb Knowledge base wrt which equivalency of patterns is checked
 * @return True if the pattern equivalent to input one is found in the trie, false otherwise 
* @throws InterruptedException 
* @throws KAON2Exception 
*/
public boolean equivalentToEarlierPattern(Pattern pattern, KnowledgeBase kb) throws KAON2Exception, InterruptedException{
	
	return depthFirstSearch(pattern, kb, tree.root());
}

/**
 * Checks if the specified pattern is not equivalent to an earlier pattern in the trie.
 *  
 * @param pattern Pattern whose equivalent is searched in the trie
 * @param kb Knowledge base wrt which equivalency of patterns is checked
 * @param parent Position of currently checked node in the trie with whom pattern is associated
 * @return True if the pattern equivalent to input one is found in the trie, false otherwise 
 * @throws InterruptedException 
 * @throws KAON2Exception 
*/

private boolean depthFirstSearch(Pattern pattern, KnowledgeBase kb, Position parent) throws KAON2Exception, InterruptedException{
	if (tree.isInternal(parent)) {
		Position child = tree.firstChild(parent);

		if (pattern.equivalentTo(((Node)child.element()).getPattern(),kb)){
			return true;
		}

		if (depthFirstSearch(pattern, kb, child)) {
			return true;
		}

		while (child != tree.lastChild(parent)) {
			child = tree.siblingAfter(child);
			if (pattern.equivalentTo(((Node)child.element()).getPattern(),kb)){
				return true;
			}
			if (depthFirstSearch(pattern, kb, child)) {
				return true;
			}
		}
	}
	return false;
}


	/**
	 * Generates dependent atoms of the given node 
	 * that is the nodes that share at least one variable that is new in the node
	 * Any semantic properties of the patterns are not tested.
	 * @param n A node that is to be refined by its dependent atoms
	 * @param bias Declarative bias
	 * @param generateDLAtoms True if DL atoms should be generated, false if only non-DL atoms should be generated 
	 * @return Nodes list (they are not added to the tree yet)
	 * @throws KAON2Exception 
	 * @throws InterruptedException 
	*/
		public ArrayList generateDependentAtoms(Node n, Bias bias, boolean generateDLAtoms) throws KAON2Exception, InterruptedException{
			ArrayList dependentAtoms = new ArrayList();
			//variables that were new in given node
			ArrayList sharedVars =  new ArrayList();
			
			//put new variables from the given node to the list of variables that may be shared with dependent node
			for (int j=0; j<n.getNewVarPos().length; j++)
			{
				if (n.getNewVarPos()[j]) {
					if (!sharedVars.contains(n.getAtom().getAtom().getArgumentVariable(j))){
						sharedVars.add(n.getAtom().getAtom().getArgumentVariable(j));
					}
				}
			}
			
			for (int i=0; i<Bias.getPredicates().size(); i++)
			{
				if (generateDLAtoms || 
						!(bias.getPredicates().get(i) instanceof OWLEntity)) {
					//generate literals with all possible variable assignments for given predicate
					ArrayList<Literal> lits = generateVariableAssignements(n.getDepth()+1, sharedVars, (Predicate) bias.getPredicates().get(i));
				
					for (Literal l : lits){
						Atom a = new Atom(l);
					
						//create refinement by adding new literal
						Pattern refinedPattern = n.getPattern().refine(a);

						//create node
						Node dependentNode = new Node();
						dependentNode.setAtom(a);
						dependentNode.setPattern(refinedPattern);
						dependentNode.setDepth(n.getDepth()+1);
						        
						boolean[] newVarPos = new boolean[l.getArity()];
						for (int j = 0; j < newVarPos.length; j++){
							if (!sharedVars.contains(l.getArgumentVariable(j))){
								newVarPos[j] = true;
							} else {
								newVarPos[j] = false;
							}	
						}
						dependentNode.setNewVarPos(newVarPos);
						dependentNode.how="a";
						        
						dependentAtoms.add(dependentNode);			        
					}
				}
			}
			return dependentAtoms;
		}
		
		/**
		 * Generates dependent atoms, with class/property predicates, of the given node 
		 * that is the nodes that share at least one variable that is new in the node
		 * Atoms are ordered wrt the taxonomy of classes/properties
		 * @param n A node that is to be refined by its dependent atoms
		 * @param bias Declarative bias
		 * @param predicateTaxonomy Taxonomy of classes/properties
		 * @param kb Knowledge base from which information about dependencies between predicates is taken
		 * @return List of dependent atoms trees/taxonomies (they are not added to the trie yet) 
		 * one tree per each variable assignment 
		 * @throws KAON2Exception 
		 * @throws InterruptedException 
		*/
		public ArrayList generateDLDependentAtoms(Node n, Bias bias, Tree predicateTaxonomy, KnowledgeBase kb) throws KAON2Exception, InterruptedException{
			ArrayList dependentAtoms = new ArrayList();
			Predicate topPredicate = (Predicate)predicateTaxonomy.root().element();
			
			//variables that were new in given node
			ArrayList sharedVars =  new ArrayList();
			
			//put new variables from the given node to the list of variables that may be shared with dependent node
			for (int j=0; j<n.getNewVarPos().length; j++){
				if (n.getNewVarPos()[j]) {
					if (!sharedVars.contains(n.getAtom().getAtom().getArgumentVariable(j))){
						sharedVars.add(n.getAtom().getAtom().getArgumentVariable(j));
					}
				}
			}
								
			ArrayList<Literal> lits = generateVariableAssignements(n.getDepth()+1, sharedVars, topPredicate);

			for (Literal l : lits){
			//copy the set of variables from top class/property literal 
			//to form atoms from each class/property in taxonomy 
	
				//add "top predicate" node as root of dependent atoms trie
				Tree da = new NodeTree();
				da.replaceElement(da.root(), l);
					
				da = recursivelyBuildDLDependentAtoms(n, l, predicateTaxonomy, predicateTaxonomy.root(), sharedVars, da, da.root());
				dependentAtoms.add(da);
			}

			return dependentAtoms;
		}
		
    /**
     * Constructs a branch of a tree of dependent atoms with class/property predicates 
     * (according to the taxonomy of classes/properties) 
     * @param n Node in patterns trie, for which dependent atoms are being constructed
     * @param l Literal to put into current node in dependent atoms tree 
     * @param t Class/property predicates taxonomy
     * @param pos Position of class/property from literal <tt>l</tt> in class/property predicates taxonomy
     * @param sharedVars Variables to be shared between atom in node <tt>n</tt> and newly created dependent atoms
     * @param dependentAtoms Dependent atoms tree
     * @param posDepAtoms Position in dependent atoms tree
	*/		
	public Tree recursivelyBuildDLDependentAtoms(Node n, Literal l, Tree t, Position pos, ArrayList sharedVars, Tree dependentAtoms, Position posDepAtoms) throws KAON2Exception {
        Position pChildDepAtoms = null;

		if (!(pos.equals(t.root()))){
			Atom a = new Atom(l);
		
			//create refinement by adding new literal
			Pattern refinedPattern = n.getPattern().refine(a);

			//create node
			Node dependentNode = new Node();
			dependentNode.setAtom(a);
			dependentNode.setPattern(refinedPattern);
			dependentNode.setDepth(n.getDepth()+1);
				        
			boolean[] newVarPos = new boolean[l.getArity()];
			for (int j = 0; j < newVarPos.length; j++){
				if (!sharedVars.contains(l.getArgumentVariable(j))){
					newVarPos[j] = true;
				} else {
					newVarPos[j] = false;
				}
			}
			dependentNode.setNewVarPos(newVarPos);
			dependentNode.how="a";
				        
			//add to dependent atoms tree
			pChildDepAtoms =  dependentAtoms.insertLastChild(posDepAtoms, dependentNode);

		} else {
			pChildDepAtoms = dependentAtoms.root();
		}
		
        for (PositionIterator pi = t.children(pos); pi.hasNext();){
        	Position pChild =  pi.nextPosition();
        	Predicate cChild = (Predicate) pChild.element();
    		Literal ll = KAON2Manager.factory().literal(true, cChild, l.getArguments());
        	dependentAtoms = recursivelyBuildDLDependentAtoms(n, ll, t, pChild, sharedVars, dependentAtoms, pChildDepAtoms);
        }
        
		return dependentAtoms;
	}
	
	
	/**
	 * Generates all possible literals with the predicate, specified as an input parameter
	 *  and variable assignments from input variables list containing:<br>
	 *  - variables specified as an input parameter (that will be shared with the parent 
	 *  of a literal)
	 *  - newly generated variables that will be new in a literal 
	 *  (in number equal to <tt>predicate_arity-1</tt>, as at least one variable in a literal should be shared) 
	 * More specifically, there are generated literals with all variations with repetitions of variables, where variations have 
	 * length equal the arity of input predicate. 
	 * The literals that do not contain any shared variable are filtered away
	 * @param sharedVars Variable list of variables that will be shared with a parent
	 * @currentDepth Depth in a trie of the parent node 
	 * @predicate predicate Predicate from which each literal is built
	 * @return All possible literals that can be built from specified (shared) and new variables and predicate
	 * and that contain at least one shared variable  
	*/

	private ArrayList generateVariableAssignements(int currentDepth, ArrayList sharedVars, Predicate predicate)
	{
		ArrayList lits = new ArrayList();
		ArrayList<Variable> allVars = new ArrayList();
		allVars.addAll(sharedVars);
		
		//Counter of new variable names
		int newVarsCounter = currentDepth*MAX_VARS;
		
		for (int i=0; i<predicate.getArity()-1;i++)
		{
			Variable v = KAON2Manager.factory().variable(String.valueOf(newVarsCounter++));
			allVars.add(v);
		}

		Variable[] vars = allVars.toArray(new Variable[allVars.size()]);

		PermutationsWithRepetition gen = new PermutationsWithRepetition(vars,predicate.getArity());
		List<Variable[]> variableAssignmentVariations = gen.getVariableVariations();
		
		for (Iterator j=variableAssignmentVariations.iterator(); j.hasNext();) {
			Variable[] variableAssignment = (Variable[]) j.next();
		
			int i=0;
			boolean containsSharedVar = false;  
			
			while (i<variableAssignment.length && !containsSharedVar) {
				if (sharedVars.contains(variableAssignment[i])) {
					containsSharedVar = true;
				}
				i++;
			}
			if (containsSharedVar) {
				Literal lit = KAON2Manager.factory().literal(true,predicate,variableAssignment);
				lits.add(lit);
			}
		}
		
		return lits;
	}
	
	
	/**
	 * Creates copy of the node, new variables are renamed to be also new in a copy.
     * @param n Node to be copied
     * @param copyParent Parent for the copy of the node
	 * @return A copy with renamed new variables and new parent attached 
	 * @throws KAON2Exception 
	*/
	public Node copy(Node n, Node copyParent) throws KAON2Exception{
		//initialize variable list for copied atom
		int[] renamedVars =  new int[n.getNewVarPos().length];
		int varNamesCounter = (n.getDepth()+1)*MAX_VARS;
		
		//rename new variables
		for (int j=0; j<renamedVars.length; j++)
		{
			renamedVars[j] = Integer.valueOf(n.getAtom().getAtom().getArgumentVariable(j).toString());

			if (n.getNewVarPos()[j]) {
				renamedVars[j] = varNamesCounter++;
			}
		}
		
		//create new node
		Term[] t = new Term[renamedVars.length];
		for (int j=0; j<renamedVars.length; j++)
		{
			Variable v = KAON2Manager.factory().variable(String.valueOf(renamedVars[j]));
			t[j]=v;
		}
		
        Literal lit=KAON2Manager.factory().literal(true,n.getAtom().getAtom().getPredicate(),t);
        Atom a = new Atom(lit);

        Pattern copyPattern = copyParent.getPattern().refine(a);
        
        Node copy = new Node();
        copy.setAtom(a);
        copy.setPattern(copyPattern);
        copy.setNewVarPos(n.getNewVarPos());
        copy.setDepth(n.getDepth()+1);
        
		return copy;
	}

	
/**
	 * Generates copies of right brothers of the given node 
	 * that is the nodes that have the same parent node as the given node 
	 * and are placed on the right of the given node.  
	 * New variables are renamed to remain new in a copy. 
	 * @param n A node that is to be refined by copies of its right brothers
	 * @param kb Knowledge base wrt which some properties (satisfibility, sfreenes) of generated refinements are tested
 * @throws KAON2Exception 
 * @throws InterruptedException 
 	*/
/*
	public void copyRightBrothers(Node n, KnowledgeBase kb) throws KAON2Exception, InterruptedException{
		Position p = n.getPos();
		if (p == tree.root()) {
			return;
		}
		
		Position parent = tree.parent(n.getPos());
		
		while (!p.equals(tree.lastChild(parent))){
            //get brother
			p = tree.siblingAfter(p);
			Node brother = (Node) p.element();
		
				//make copy of brother
				Node brotherCopy = copy(brother, n);
				brotherCopy.how="b";
		
				//check whether pattern extended by new literal is satisfiable
				if (brotherCopy.getPattern().isSatisfiable(kb)) {
					//check whether pattern is s-free
					if (brotherCopy.getPattern().isSFree(kb)) {
						//check whether pattern is not equivalent to earlier pattern in trie
						if (!equivalentToEarlierPattern(brotherCopy.getPattern(), kb)) {
            
							//add new node (copy) to the tree
							addNode(n, brotherCopy);
						}
					}
				}
		}
	}
*/

	/**
	 * Generates copies of right brothers of the given node 
	 * that is the nodes that have the same parent node as the given node 
	 * and are placed on the right of the given node.  
	 * New variables are renamed to remain new in a copy.
	 * Any semantic properties of the patterns are not tested.
	 * @param n A node that is to be refined by copies of its right brothers
	 * @return Nodes list (they are not added to the tree yet)
 * @throws KAON2Exception 
 * @throws InterruptedException 
 	*/
	public ArrayList generateRightBrothers(Node n) throws KAON2Exception, InterruptedException{
		ArrayList brothers = new ArrayList();
		Position p = n.getPos();
		if (p == tree.root()) {
			return brothers;
		}
		
		Position parent = tree.parent(n.getPos());
		
		while (!p.equals(tree.lastChild(parent))){
            //get brother
			p = tree.siblingAfter(p);
			Node brother = (Node) p.element();
		
				//make copy of brother
				Node brotherCopy = copy(brother, n);
				brotherCopy.how="b";
		
				//check whether pattern extended by new literal is satisfiable
//				if (brotherCopy.getPattern().isSatisfiable(kb)) {
					//check whether pattern is s-free
//					if (brotherCopy.getPattern().isSFree(kb)) {
						//check whether pattern is not equivalent to earlier pattern in trie
//						if (!equivalentToEarlierPattern(brotherCopy.getPattern(), kb)) {
            
							//add new node (copy) to the tree
				brothers.add(brotherCopy);
//							addNode(n, brotherCopy);
//						}
//					}
//				}
		}
		return brothers;
	}

	
	/**
	 * Generates copies of the given node 
	 * @param n A node that is to be refined by its copies
	 * @param kb Knowledge base wrt which some properties (satisfibility, sfreenes) of generated refinements are tested
	 * @throws KAON2Exception 
	 * @throws InterruptedException 
	*/
	public void copyItself(Node n, KnowledgeBase kb) throws KAON2Exception, InterruptedException{
		Position p = n.getPos();
		if (p == tree.root()) {
			return;
		}
		else if (n.getAtom().getAtom().getArity() == 1) {
			return;
		}

		//make copy of the node
        Node copy = copy(n, n);
        copy.how="c";

		//check whether pattern extended by new literal is satisfiable
		if (copy.getPattern().isSatisfiable(kb)) {
			//check whether pattern is s-free
			if (copy.getPattern().isSFree(kb)) {
				//check whether pattern is not equivalent to earlier pattern in trie
				if (!equivalentToEarlierPattern(copy.getPattern(), kb)) {
        
					//add new node (copy) to the tree
					addNode(n, copy);
				}
			}
		}
	}

	/**
	 * Adds specified node as child to the specified parent node 
	 * @param parent Parent node
	 * @param child Child node 
	*/
	public void addNode(Node parent, Node child) {
		Position posChild = tree.insertLastChild(parent.getPos(), child);
		child.setPos(posChild);
		HashMap lmd = new HashMap();
		lmd.putAll(parent.getLastMaterializationDepth());
		child.setLastMaterializationDepth(lmd);
		tree.replaceElement(child.getPos(), child);
	}

	/**
	 * Removes specified node and optionally deletes its associated resources 
	 * @param node Node to remove
	 * @param disposeAssociatedResources True if the resources associted to given node, e.g. query,
	 * should be also disposed while node is being removed
	 **/
	public void removeNode(Node node, boolean disposeAssociatedResources) {
		if (disposeAssociatedResources) {
			//node.getPattern().getQuery().dispose();
			node.setPattern(null);
			
			if (node.getRewrittenPattern()!=null){
				//System.out.println("usuwa rewritten");
				//node.getRewrittenPattern().getQuery().dispose();
				node.setRewrittenPattern(null);
			}
		}
	 	tree.removeExternal(node.getPos());
	}
	 	
	/**
	 * Expands given node by generating its children as:
	 *  dependent atoms, right brothers and node copies.
	 * @param n A node to be expanded 
	 * @param kb Knowledge base wrt which some properties (satisfibility, sfreenes) of generated refinements are tested
	 * @param bias Declarative bias  
	 * @throws InterruptedException 
	 * @throws KAON2Exception 
	*/
//	public void expand(Node n, KnowledgeBase kb, Bias bias) throws KAON2Exception, InterruptedException{
//		generateDependentAtoms(n, kb, bias);
//		copyRightBrothers(n, kb);
		//copyItself(n, kb);
//	}
	
	/**
	 * Testing Trie class methods
	*/
	public static void main(String[] args) {
		Variable v1=KAON2Manager.factory().variable("x");
		Variable v2=KAON2Manager.factory().variable("y");	
		ArrayList sharedVars = new ArrayList();
		sharedVars.add(v1);
		sharedVars.add(v2);

		Node n = new Node();
		
		ObjectProperty isOwnerOf=KAON2Manager.factory().objectProperty("http://www.owl-ontologies.com/unnamed.owl#isOwnerOf");
		Trie trie = new Trie(n);
		trie.generateVariableAssignements(1, sharedVars, isOwnerOf);
	}

}
