/**
 * Created on 2008-03-18 11:22:05
 */
package org.put.semintec.mining.association;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import jdsl.core.api.Position;
import jdsl.core.api.PositionIterator;
import org.put.semintec.core.Atom;
import org.put.semintec.core.KnowledgeBase;
import org.put.semintec.mining.datastructures.Node;
import org.put.semintec.mining.datastructures.Trie;
import org.put.semintec.utils.Statistics;
import org.semanticweb.kaon2.api.KAON2Exception;
import org.semanticweb.kaon2.api.KAON2Manager;
import org.semanticweb.kaon2.api.OntologyChangeEvent;
import org.semanticweb.kaon2.api.formatting.OntologyFileFormat;
import org.semanticweb.kaon2.api.logic.Formula;
import org.semanticweb.kaon2.api.logic.Literal;
import org.semanticweb.kaon2.api.logic.QueryDefinition;
import org.semanticweb.kaon2.api.logic.Term;
import org.semanticweb.kaon2.api.logic.Variable;
import org.semanticweb.kaon2.api.owl.elements.OWLClass;
import org.semanticweb.kaon2.api.reasoner.Query;

/**
 * Represents an algorithm for mining patterns that do not exploit KB semantics during pattern generation.
 * @author Agnieszka Lawrynowicz
 * 
 */
public class PatternMiningWithoutBackgroundKnowledge {

	
/**
 * List of maximally general infrequent patterns.
 */
private ArrayList infrequentPatterns;
 
/**
 * Getter of the property <tt>infrequentPatterns</tt>
 *
 * @return Returns the infrequentPatterns.
 * 
 */
public ArrayList getInfrequentPatterns()
{
	return infrequentPatterns;
}

/**
 * Setter of the property <tt>infrequentPatterns</tt>
 *
 * @param infrequentPatterns The infrequentPatterns to set.
 *
 */
public void setInfrequentPatterns(ArrayList infrequentPatterns ){
	this.infrequentPatterns = infrequentPatterns;
}


/**
 * Declarative bias
 */
private Bias bias;
 
/**
 * Getter of the property <tt>bias</tt>
 *
 * @return Returns the bias.
 * 
 */
public Bias getBias()
{
	return bias;
}

/**
 * Setter of the property <tt>bias</tt>
 *
 * @param bias The bias to set.
 *
 */
public void setBias(Bias bias ){
	this.bias = bias;
}

/**
 * Stores statistics of pattern mining execution
 */
private Statistics statistics;
 
/**
 * Getter of the property <tt>statistics</tt>
 *
 * @return Returns the statistics.
 * 
 */
public Statistics getStatistics()
{
	return statistics;
}

/**
 * Setter of the property <tt>statistics</tt>
 *
 * @param statistics The statistics to set.
 *
 */
public void setStatistics(Statistics statistics ){
	this.statistics = statistics;
}

/**
 * Depth-first search direction for patterns.   
 */
public static int DEPTH_FIRST_SEARCH = 0;	

/**
 * Breadth-first search direction for patterns.   
 */
public static int BREADTH_FIRST_SEARCH = 1;	


/**
 * Search direction for patterns.
 */
private int searchDirection = DEPTH_FIRST_SEARCH;
 
/**
 * Getter of the property <tt>searchDirection</tt>
 *
 * @return Returns the searchDirection.
 * 
 */
public int getSearchDirection()
{
	return searchDirection;
}

/**
 * Setter of the property <tt>searchDirection</tt>
 *
 * @param searchDirection The searchDirection to set.
 *
 */
public void setSearchDirection(int searchDirection ){
	this.searchDirection = searchDirection;
}

/**
 * Flag for indicating if answer caching mechanism with query rewriting based on cached answers should be on 
 */
private boolean answerCaching = false;
 
/**
 * Getter of the property <tt>answerCaching</tt>
 *
 * @return Returns the answerCaching.
 * 
 */
public boolean getAnswerCaching()
{
	return answerCaching;
}

/**
 * Setter of the property <tt>answerCaching</tt>
 *
 * @param answerCaching The answerCaching to set.
 *
 */
public void setAnswerCaching(boolean answerCaching ){
	this.answerCaching = answerCaching;
}

/**
 * Trie data structure on which mining algorithm operates.
 */
private Trie trie;
 
/**
 * Getter of the property <tt>trie</tt>
 *
 * @return Returns the trie.
 * 
 */
public Trie getTrie()
{
	return trie;
}

/**
 * Setter of the property <tt>trie</tt>
 *
 * @param trie The trie to set.
 *
 */
public void setTrie(Trie trie ){
	this.trie = trie;
}

/**
 * Knowledge base on which mining algorithm is run
 */
private KnowledgeBase kb;
 
/**
 * Getter of the property <tt>kb</tt>
 *
 * @return Returns the kb.
 * 
 */
public KnowledgeBase getKb()
{
	return kb;
}

/**
 * Setter of the property <tt>kb</tt>
 *
 * @param kb The kb to set.
 *
 */
public void setKb(KnowledgeBase kb ){
	this.kb = kb;
}

/**
 * Copy of the knowledge base on which mining algorithm is run.
 * Used in pattern satisfiability and containment tests
 */
private KnowledgeBase kbCopy;

/**
 * Setter of the property <tt>kbCopy</tt>
 *
 * @param kbCopy The kbCopy to set.
 *
 */
public void setKbCopy(KnowledgeBase kbCopy ){
	this.kbCopy = kbCopy;
}

 
/**
 * Getter of the property <tt>kbCopy</tt>
 *
 * @return Returns the kbCopy.
 * 
 */
public KnowledgeBase getKbCopy()
{
	return kbCopy;
}



/**
 * Variable in an atom with reference class as predicate.
 */
private Variable refVar;
 
/**
 * Getter of the property <tt>refVar</tt>
 *
 * @return Returns the refVar.
 * 
 */
public Variable getRefVar()
{
	return refVar;
}

/**
 * Setter of the property <tt>refVar</tt>
 *
 * @param refVar The refVar to set.
 *
 */
public void setRefVar(Variable refVar ){
	this.refVar = refVar;
}

/**
 * Reference class, with regards to which support is calculated and pattern discovery run.   
 */
private OWLClass refClass;
 
/**
 * Getter of the property <tt>refClass</tt>
 *
 * @return Returns the refClass.
 * 
 */
public OWLClass getRefClass()
{
	return refClass;
}

/**
 * Setter of the property <tt>refClass</tt>
 *
 * @param refClass The refClass to set.
 *
 */

public void setRefClass(OWLClass refClass ){
	this.refClass = refClass;
}

/**
 * Reference pattern - trivial one   
 */
private Pattern refPattern;
 
/**
 * Getter of the property <tt>refPattern</tt>
 *
 * @return Returns the refClass.
 * 
 */
private Pattern getRefPattern()
{
	return refPattern;
}

/**
 * Setter of the property <tt>refPattern</tt>
 *
 * @param refPattern The refPattern to set.
 *
 */
private void setRefPattern(Pattern refPattern ){
	this.refPattern = refPattern;
}


/**
 * Maximal length (number of atoms) of a discovered pattern. 
 */
private int maxLength;
 
/**
 * Getter of the property <tt>maxLength</tt>
 *
 * @return Returns the maxLength.
 * 
 */
public int getMaxLength()
{
	return maxLength;
}

/**
 * Setter of the property <tt>maxLength</tt>
 *
 * @param maxLength The maxLength to set.
 *
 */
public void setMaxLength(int maxLength ){
	this.maxLength = maxLength;
}

/**
 * Minimum support threshold specified by the user.
 */
private double minSupport;
 
/**
 * Getter of the property <tt>minSupport</tt>
 *
 * @return Returns the minSupport.
 * 
 */
public double getMinSupport()
{
	return minSupport;
}

/**
 * Setter of the property <tt>minSupport</tt>
 *
 * @param minSupport The minSupport to set.
 *
 */
public void setMinSupport(double minSupport ){
	this.minSupport = minSupport;
}


/**
	 * Executes the algorithm for pattern mining. 
	*/
	public void mine(){
		try {
			initialize();
			if (getSearchDirection() == DEPTH_FIRST_SEARCH) {
				depthFirstExpand((Node)trie.getTree().root().element());
			} else if (getSearchDirection() == BREADTH_FIRST_SEARCH) {
				//TODO breadth-first search
		}

		} catch (KAON2Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	
	/**
	 * Mines for patterns expanding the trie depth first 
	 *  
	 * @param n Current node to expand
	 * @throws InterruptedException 
	 * @throws KAON2Exception 
	*/

	private void depthFirstExpand(Node n) throws KAON2Exception, InterruptedException{
		Position parent = n.getPos();
		if (n.getDepth()<getMaxLength()) {
			//System.gc();

			List<OntologyChangeEvent> rollback = null;
		
		//find all children of n
	        ArrayList generatedChildren = new ArrayList();
	        generatedChildren.addAll(trie.generateDependentAtoms(n, bias,true));
	        generatedChildren.addAll(trie.generateRightBrothers(n));
	        
	     //   System.out.println("ch_size:"+generatedChildren.size());
	        
		//test semantically and evaluate all children of n
			
			for (Iterator i = generatedChildren.iterator(); i.hasNext();) {
				//take next child
				Node nChild = (Node)i.next();
	//			System.out.println("ch:"+nChild.getPattern().getQueryDefinition().toString());
				
				//check whether pattern extended by new literal is satisfiable
//		    	if (nChild.getPattern().isSatisfiable(kbCopy)) {
		    		//check whether pattern is s-free
//		    		if (nChild.getPattern().isSFree(kbCopy)) {
		    			//check whether pattern is not subsumed by already found maximally general infrequent one
		    		//	if (!subsumedByInfrequentPattern(nChild.getPattern(),kbCopy)) {
		    				//check whether pattern is not equivalent to earlier pattern in trie
//		   					if (!trie.equivalentToEarlierPattern(nChild.getPattern(),kbCopy)) {
		    					//collect statistics
		    					getStatistics().setNumCandidatesUpToMaxLength(getStatistics().getNumCandidatesUpToMaxLength()+1);
		    					if (nChild.getDepth()==getMaxLength()) {
		    						getStatistics().setNumCandidatesWithMaxLength(getStatistics().getNumCandidatesWithMaxLength()+1);
		    					}
			
		    					//if there is any child, materialize given node before evaluating and expanding it
		    					if (getAnswerCaching() && !n.getMaterialized()){
		    						if (n.getDepth()==1){
		    							rollback=n.materializeResultsInit(kb);
		    						} else {
		    							rollback=n.materializeResults(kb);
		    						}
		    						n.setMaterialized(true);
		    					}

		    					Pattern p;
		    					if (!getAnswerCaching()){
		    						p = nChild.getPattern(); 
		    					} else {
		    						p = nChild.rewriteLiteralsToSupOnes(n, kb);
		    						//nChild.setRewrittenPattern(p);
		    					}
		    					
		    					p.evaluate(refPattern.getFrequency(), kb);
		    					
		    					if (!p.isFrequent(getMinSupport())){
		    				//		//put to infrequent patterns list
		    					//	if (!addToInfrequentPatterns(nChild.getPattern(), kbCopy)) {
		    							nChild.dispose();
		    							nChild=null;
		    					//	}
		    					} else {
		    						if (getAnswerCaching()) {
		    							nChild.getPattern().setSupport(p.getSupport());
		    							nChild.getPattern().setFrequency(p.getFrequency());
		    						}
					
		    						trie.addNode(n, nChild);
						
		    						//collect statistics
		    						getStatistics().setNumPatternsUpToMaxLength(getStatistics().getNumPatternsUpToMaxLength()+1);
		    						if (nChild.getDepth()==getMaxLength()) {
		    							getStatistics().setNumPatternsWithMaxLength(getStatistics().getNumPatternsWithMaxLength()+1);
		    						}
		    					}
		    				}
		//    			}
					//}
	//	    	}
	//		}
		
		//refine recursively each child of n
			//if there are any children of parent node after evaluating candidates
 			//and if children depth is not MAX_DEPTH 
			if (trie.getTree().isInternal(parent) && n.getDepth()<getMaxLength()-1){
				Position child = trie.getTree().firstChild(parent);

				depthFirstExpand((Node) child.element());
			
				while (child != trie.getTree().lastChild(parent)) {
					child = trie.getTree().siblingAfter(child);
					depthFirstExpand((Node) child.element());
				}
			}
			
			//rollback materializations
			if (getAnswerCaching() && n.getMaterialized()){
				kb.getOntology().applyChanges(rollback);
			}
		}
	}
	
	/**
	 * Initializes data for mining algorithm 
	 * @throws KAON2Exception 
	 * @throws InterruptedException 
	*/
	private void initialize() throws KAON2Exception, InterruptedException {
		//reference variable
		refVar = KAON2Manager.factory().variable("-1");
		
		//reference atom
		Literal refAtom = KAON2Manager.factory().literal(true,refClass,new Term[] { refVar });
		
		//root node of the trie
		Node rootNode = new Node();
		rootNode.setAtom(new Atom(refAtom));
		
		boolean[] newVar = new boolean[1];
		newVar[0]=true;
		rootNode.setNewVarPos(newVar);
		
		Formula f = KAON2Manager.factory().conjunction(refAtom);
		
		QueryDefinition refQueryDefinition = KAON2Manager.factory().queryDefinition(f, new Variable[]{refVar});
			//			kb.getReasoner().createQuery(new Formula[]{refAtom},new Variable[]{refVar});
		refPattern = new Pattern(refQueryDefinition);
		
		Query refQuery = kb.getReasoner().createQuery(refQueryDefinition);
		//calculate frequency of reference pattern
		refQuery.open();
		refPattern.setFrequency(refQuery.getNumberOfTuples());
		System.out.println("reffreq:"+refPattern.getFrequency());
		refQuery.close();
		refQuery.dispose();

		refPattern.setSupport(1.0);

		rootNode.setPattern(refPattern);
		rootNode.setDepth(1);

		
	//	if (getAnswerCaching()) {
			//Pattern rewrittenPattern = new Pattern(refQueryDefinition);
			//rewrittenPattern.setFrequency(refPattern.getFrequency());
			//rewrittenPattern.setSupport(1.0);
			//rewrittenPattern.computeUndistinguishedVariables();
			//rootNode.setRewrittenPattern(rewrittenPattern);
			//rootNode.materializeResultsInit(kb);
	//	}
	
		infrequentPatterns = new ArrayList();
		
		//initialize trie with the root node
		trie = new Trie(rootNode);		
	}

	
	/**
	 * @param p A pattern that is tested whether it is not subsumed by already found infrequent one
	 * @param kb A knowledge base w.r.t. pattern subsumption is tested. 
	 * @return True if the given pattern is subsumed by already found infrequent one w.r.t. given knowledge base   
	 * @throws InterruptedException 
	 * @throws KAON2Exception 
	*/
	public boolean subsumedByInfrequentPattern(Pattern p, KnowledgeBase kb) throws KAON2Exception, InterruptedException {
		int i = 0;
		
		while (i<getInfrequentPatterns().size()) {
			if (p.subsumedBy((Pattern)getInfrequentPatterns().get(i), kb)) {
				return true;
			}
			i++;
		}
		return false;
	}

	/**
	 * @param p A pattern to be added to infrequent patterns list. 
	 * In order to keep at the list the patterns that are maximally general, it is tested if the pattern is not
	 * subsumed by some pattern already present in the list and if this pattern does not subsume any other pattern on the list 
	 * @param kb A knowledge base w.r.t. pattern subsumption is tested.
	 * @return True if pttern was added to infrequent patterns list, false otherwise (more general pattern has been found there)
	 * @throws InterruptedException 
	 * @throws KAON2Exception 
	*/
	public boolean addToInfrequentPatterns(Pattern p, KnowledgeBase kb) throws KAON2Exception, InterruptedException {
		int i = 0;
		boolean foundMoreGeneral = false;
		ArrayList updatedInfrequentPatternsList = new ArrayList();
		
		while (i<getInfrequentPatterns().size() && !foundMoreGeneral) {
			if (p.subsumedBy((Pattern)getInfrequentPatterns().get(i), kb)) {
				foundMoreGeneral = true;
			}
			i++;
		}
		if (!foundMoreGeneral) {

			i = 0;
			while (i<getInfrequentPatterns().size()) {
				if (!((Pattern)getInfrequentPatterns().get(i)).subsumedBy(p, kb)) {
					updatedInfrequentPatternsList.add((Pattern)getInfrequentPatterns().get(i));
				} 
				
				i++;
			}
			
			updatedInfrequentPatternsList.add(p);
			setInfrequentPatterns(new ArrayList());
			getInfrequentPatterns().addAll(updatedInfrequentPatternsList);
			
			return true;
		} else {
			return false;
		}
	}
	
}

